/*
*
*  * Copyright (c) 2019. Honeywell International, Inc. All Rights Reserved.
*
*/
package com.honeywell.easytemplating.util;

import com.honeywell.easytemplating.enums.BEasyTemplatingToastEnum;
import com.tridium.install.BDependency;
import com.tridium.install.BVersion;
import com.tridium.template.BPasswordBinding;
import com.tridium.template.BRelationInfo;
import com.tridium.template.BTemplateConfig;
import com.tridium.template.BTemplateState;
import com.tridium.template.UpgradeUtil;
import com.tridium.template.file.BNtplFile;
import com.tridium.template.file.DependencyUtil;
import com.tridium.template.file.PxFileRef;
import com.tridium.template.manifest.ManifestXMLWriter;
import com.tridium.template.manifest.TemplateManifest;
import com.tridium.template.ui.BTemplateOptions;
import com.tridium.template.ui.BTemplateView;
import com.tridium.template.ui.file.BWbDeployableNtplFile;
import com.tridium.template.ui.file.TmplUtil;
import com.tridium.util.CompUtil;
import com.tridium.util.ObjectUtil;

import javax.baja.agent.BPxView;
import javax.baja.file.BDirectory;
import javax.baja.file.BFileSystem;
import javax.baja.file.BIFile;
import javax.baja.file.BajaFileUtil;
import javax.baja.file.FilePath;
import javax.baja.file.types.image.BImageFile;
import javax.baja.file.types.text.BPxFile;
import javax.baja.gx.BImage;
import javax.baja.io.ValueDocEncoder;
import javax.baja.naming.BOrd;
import javax.baja.naming.SlotPath;
import javax.baja.naming.UnresolvedException;
import javax.baja.nre.platform.RuntimeProfile;
import javax.baja.sys.BAbsTime;
import javax.baja.sys.BComponent;
import javax.baja.sys.BFacets;
import javax.baja.sys.BInteger;
import javax.baja.sys.BMarker;
import javax.baja.sys.BString;
import javax.baja.sys.Flags;
import javax.baja.sys.Property;
import javax.baja.sys.Sys;
import javax.baja.util.BUuid;
import javax.baja.util.Lexicon;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Locale;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_ERROR_INMEM_NTPL;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_ERROR_NTPL_IO;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_ERROR_ORD_RESOLUTION;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_INFO_TAG_NAME;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_MAIN_VIEW;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_NTPL_FILE_TAG_NAME;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_OK_NTPL_CREATION;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_OK_NTPL_CREATION_W_PARAM;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_POPUP_BINDING_PREFIX;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_POPUP_VIEW;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_READONLY_HIDDEN_META;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_ROOT_TAG_NAME;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_TEMPLATE_FOLDER;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_TEMPLATE_FOLDER_ORD;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_TEMPLATE_VENDOR;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_TITLE_TAG_NAME;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_TOAST_ICON_ORD_OK;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_TOAST_ICON_ORD_WARNING;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_TOAST_LONG_TIMEOUT;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_UID_TAG_NAME;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_VENDOR_TAG_NAME;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_VERSION_TAG_NAME;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.TEMPLATE_RELATION;

public final class EasyTemplatingGenerationHelper {

    private EasyTemplatingGenerationHelper() {
    }

    //***********************************************************************************************
    // PUBLIC API
    //***********************************************************************************************
    private static final Lexicon lex = Lexicon.make(EasyTemplatingGenerationHelper.class);
    private static void isRootHasView(BComponent root) throws IllegalArgumentException {

        Property[] properties = root.getDynamicPropertiesArray();

        for (Property property : properties) {
            if (property.getType().is(BPxView.TYPE)) {
                throw new IllegalArgumentException("Please remove the view from " + root.getSlotPathOrd());
            }
        }
    }

    public static void generateTemplate(
            BComponent rootComponent,
            String templateName,
            String neqlQuery,
            BOrd objectPxOrd,
            BOrd popupPxOrd) throws IllegalArgumentException, EasyTemplatingBuilderException {


        isRootHasView(rootComponent);

        BWbDeployableNtplFile templateFileNoViews = makeMemoryBasedTemplateFile(rootComponent);
        if (null == templateFileNoViews) {
            return;
        }

        BComponent templateComponent = templateFileNoViews.getBaseComponent();
        BComponent templateComponentParent = templateComponent.getParent().asComponent();

        EasyTemplatingUtils.addLinkerToComponent(templateComponent);
        EasyTemplatingUtils.addDynamicViewToTheComponent(templateComponent, objectPxOrd, ET_MAIN_VIEW);
        if (!popupPxOrd.isNull()) {
            EasyTemplatingUtils.addDynamicViewToTheComponent(templateComponent, popupPxOrd, ET_POPUP_VIEW);
        }

        // make new template file with component with views on
        BWbDeployableNtplFile templateFile = makeMemoryBasedTemplateFile(templateComponent);
        if (null == templateFile) {
            return;
        }

        TemplateManifest manifest = templateFile.getTemplateManifest();

        List<PxFileRef> memoryBasedPxFiles = new ArrayList<>();
        //Update manifest with all px files
        memoryBasedPxFiles= Arrays.asList(templateFile.getPxFiles());


        BTemplateConfig templateConfig = ensureTemplateConfigFromComponent(templateComponent);

        if (manifest.newCreation) {
            templateName = processNewTemplate(manifest, templateConfig, templateComponent, templateName, neqlQuery);
        } else {
            templateName = templateConfig.getTemplateName();
            if (manifest.title.isEmpty()) {
                manifest.title = templateName;
            }
            if (templateConfig.getUID().equals(BUuid.DEFAULT)) {
                templateConfig.setUID(BUuid.make());
                manifest.uID = templateConfig.getUID();
            }
        }

        Property templateConfigProperty = templateConfig.getPropertyInParent();
        BFacets f = templateComponent.getSlotFacets(templateConfigProperty);
        templateComponent.setFacets(templateConfigProperty, BFacets.makeRemove(f, "ntplCreation"));

        templateConfig.setTemplateName(templateName);
        templateConfig.setVersion(makeVersionFromManifest(manifest));
        templateConfig.setVersionDate(BAbsTime.now());
        if (!templateConfig.getUID().equals(manifest.uID)) {
            manifest.uID = templateConfig.getUID();
        }

        TemplateManifest targetManifest = templateConfig.getManifest();
        setTargetManifestPropsFromFile(targetManifest, templateFile);
        ensureTemplateComponentProps(templateComponent, targetManifest, templateFile);

        //build a new list to add all memory based px file to a dependencies
        int pxFileRefsLenght = templateFile.getPxFiles().length + memoryBasedPxFiles.size();
        PxFileRef[] allPxFiles = new PxFileRef[pxFileRefsLenght];
        buildMemoryPxFileArray(templateFile, allPxFiles, memoryBasedPxFiles);
        addModuleDependenciesToManifest(targetManifest, templateComponent, allPxFiles);
        createNewTemplateFile(templateName, templateComponentParent, templateConfig, templateFile, allPxFiles);
    }


    public static BDirectory getTemplateDirectory() {
        try {
            return (BDirectory) getTemplateDirectoryOrd().resolve().get();
        } catch (Exception e) {
            EasyTemplatingLogger.error(e.getLocalizedMessage(), e);
            return null;
        }
    }

    public static BOrd getTemplateDirectoryOrd() {
        BOrd tplOrd = BOrd.make(ET_TEMPLATE_FOLDER_ORD);
        try {
            tplOrd.resolve();
        } catch (UnresolvedException ue) {
            EasyTemplatingLogger.error(String.format(ET_ERROR_ORD_RESOLUTION, ET_TEMPLATE_FOLDER_ORD), ue);
        }
        return tplOrd;
    }

    //***********************************************************************************************
    // PRIVATE API
    //***********************************************************************************************

    /**
     * Build pxFileRef [] with memory based px files
     *
     * @param templateFile
     * @param allPxFiles
     * @param memoryBasedPxFiles
     */
    private static void buildMemoryPxFileArray(BWbDeployableNtplFile templateFile,
                                               PxFileRef[] allPxFiles,
                                               List<PxFileRef> memoryBasedPxFiles) {
        //Add template px file references
        PxFileRef[] templatePxFiles = templateFile.getPxFiles();
        for (int i = 0; i < templateFile.getPxFiles().length; ++i) {
            memoryBasedPxFiles.add(templatePxFiles[i]);
        }
        //Copy all file to array
        memoryBasedPxFiles.toArray(allPxFiles);
    }

    /*
     * In a future sprint, UI feedback will be provided.
     * For now, the templateName variable is unused.
     */
    private static BWbDeployableNtplFile makeMemoryBasedTemplateFile(BComponent rootComponent) {
        BWbDeployableNtplFile templateFile = null;
        try {
            templateFile = (BWbDeployableNtplFile) TmplUtil.createInMemoryNtpl(rootComponent);
        } catch (Exception e) {
            BEasyTemplatingToast.raise(
                    null,
                    BImage.make(ET_TOAST_ICON_ORD_WARNING),
                    BString.make(lex.getText("EasyTemplate.Create.InMemory.Template.File")),
                    BString.make(lex.getText("EasyTemplate.Component.Name" , new Object[]{SlotPath.unescape(rootComponent.getName())})),
                    BInteger.make(ET_TOAST_LONG_TIMEOUT),
                    BEasyTemplatingToastEnum.make(BEasyTemplatingToastEnum.ERROR));
            EasyTemplatingLogger.error(ET_ERROR_INMEM_NTPL, e);
        }
        return templateFile;
    }

    private static void setTargetManifestPropsFromFile(TemplateManifest targetManifest, BNtplFile ntplFile) {
        TemplateManifest fileManifest = ntplFile.getTemplateManifest();
        if (null != targetManifest && null != fileManifest) {
            targetManifest.resources = fileManifest.resources;
            targetManifest.version = fileManifest.version;
            targetManifest.bogSignature = fileManifest.bogSignature;
            targetManifest.title = fileManifest.title;
            targetManifest.buildVersion =
                    Sys
                            .getModuleForClass(BTemplateView.class)
                            .getVendorVersion(RuntimeProfile.rt)
                            .toString();
        }
    }

    private static BTemplateConfig ensureTemplateConfigFromComponent(BComponent aComponent) {
        BTemplateConfig templateConfig = null;
        Property templateConfigProperty = aComponent.getProperty("templateConfig");
        if (templateConfigProperty == null) {
            templateConfig = new BTemplateConfig();
            aComponent.add(
                    "templateConfig",
                    templateConfig,
                    0,
                    BFacets.make("ntplCreation", true),
                    null);
        } else {
            templateConfig = ((BTemplateConfig) aComponent.get(templateConfigProperty));
        }
        return templateConfig;
    }

    private static void ensureTemplateComponentProps(BComponent templateComponent,
                                                     TemplateManifest targetManifest,
                                                     BNtplFile ntplFile) {
        CompUtil.setOrAdd(templateComponent, ET_ROOT_TAG_NAME, BMarker.MARKER,
                ET_READONLY_HIDDEN_META, null, null);
        CompUtil.setOrAdd(templateComponent, ET_NTPL_FILE_TAG_NAME,
                BString.make(ntplFile.getFileName()),
                ET_READONLY_HIDDEN_META, null, null);
        CompUtil.setOrAdd(templateComponent, ET_VENDOR_TAG_NAME,
                BString.make(targetManifest.vendor), ET_READONLY_HIDDEN_META, null, null);
        CompUtil.setOrAdd(templateComponent, ET_TITLE_TAG_NAME,
                BString.make(targetManifest.title), ET_READONLY_HIDDEN_META, null, null);
        CompUtil.setOrAdd(templateComponent, ET_VERSION_TAG_NAME,
                BString.make(targetManifest.version), ET_READONLY_HIDDEN_META, null, null);
        CompUtil.setOrAdd(templateComponent, ET_INFO_TAG_NAME,
                BString.make(targetManifest.description), ET_READONLY_HIDDEN_META, null, null);
        CompUtil.setOrAdd(templateComponent, ET_UID_TAG_NAME,
                BString.make(targetManifest.uID.toString()), ET_READONLY_HIDDEN_META, null, null);
    }

    private static BVersion makeVersionFromManifest(TemplateManifest templateManifest) {
        BVersion version = new BVersion("Tridium", templateManifest.version);
        version.setBajaVersionString(Sys.getBajaVersion().toString());
        return version;
    }

    private static void addModuleDependenciesToManifest(TemplateManifest templateManifest,
                                                        BComponent rootComponent,
                                                        PxFileRef[] pxFileRefs) {
        EasyTemplatingLogger.info("Adding module dependencies to template config manifest...");
        Hashtable<String, BDependency> hashtable = new Hashtable();

        DependencyUtil.getBogComponentDependencies(
                rootComponent,
                hashtable,
                BTemplateOptions.get().getUseMinorVersionOnDeployment());

        DependencyUtil.getTemplatePxDependencies(
                pxFileRefs,
                hashtable,
                BTemplateOptions.get().getUseMinorVersionOnDeployment());

        BDependency[] dependencies = new BDependency[hashtable.size()];
        templateManifest.addDependencies(hashtable.values().toArray(dependencies));
    }

    /**
     * @param manifest
     * @param templateConfig
     * @param templateComp
     * @return Template name
     */
    private static String processNewTemplate(
            TemplateManifest manifest,
            BTemplateConfig templateConfig,
            BComponent templateComp,
            String templateName,
            String neql) {
        BDirectory templates = EasyTemplatingGenerationHelper.getTemplateDirectory();
        templateName = EasyTemplatingGenerationHelper.getUniqueName(templateName, templates);
        manifest.title = templateName;
        manifest.vendor = ET_TEMPLATE_VENDOR;
        manifest.version = "1.0";
        BPasswordBinding[] passwordBindings = templateConfig.getPasswordBindings();
        for (BPasswordBinding binding : passwordBindings) {
            if (binding.getIsDynamic()) {
                templateConfig.remove(binding);
            }
        }
        manifest.state = BTemplateState.DEFAULT;
        manifest.uID = BUuid.make();
        templateConfig.setUID(manifest.uID); // Is this required ?
        templateConfig.add(
                "relate?",
                BRelationInfo.make(
                        true,
                        TEMPLATE_RELATION.toString(),
                        SlotPath.unescape(neql),
                        ""),
                Flags.READONLY + Flags.HIDDEN,
                null);
        long signature = UpgradeUtil.getTemplateSignature(templateComp);
        manifest.bogSignature = Long.toHexString(signature);
        return templateName;
    }

    private static void createNewTemplateFile(
            String templateName,
            BComponent root,
            BTemplateConfig config,
            BNtplFile ntplFile,
            PxFileRef[] memoryPxFiles) {
        try {
            makeTemplateFile(templateName, root, config, null, ntplFile, true, memoryPxFiles);
            BEasyTemplatingToast.raise(
                    null,
                    BImage.make(ET_TOAST_ICON_ORD_OK),
                    BString.make(ET_OK_NTPL_CREATION),
                    BString.make(templateName + ".ntpl"),
                    null,
                    BEasyTemplatingToastEnum.make(BEasyTemplatingToastEnum.SUCCESS));
        } catch (IOException ioe) {
            EasyTemplatingLogger.error(ET_ERROR_NTPL_IO, ioe);
            BEasyTemplatingToast.raise(
                    null,
                    BImage.make(ET_TOAST_ICON_ORD_WARNING),
                    BString.make(ET_ERROR_NTPL_IO),
                    BString.make(templateName),
                    BInteger.make(ET_TOAST_LONG_TIMEOUT),
                    BEasyTemplatingToastEnum.make(BEasyTemplatingToastEnum.ERROR));
        }
    }

    /**
     * If image required, pass in something similar to:
     * ((BImageFile)BOrd.make(ET_TEMPLATE_ICON_ORD)).get(templateComp)
     */
    private static void makeTemplateFile(
            String templateName,
            BComponent root,
            BTemplateConfig config,
            BImageFile image,
            BNtplFile templateFile,
            boolean isNew,
            PxFileRef[] memoryPxFiles) throws IOException {

        BImageFile[] pxImages = null;
        PxFileRef[] pxFiles = null;
        BIFile biFile = null;
        String outFilename = null;

        if (templateFile != null) {
            pxImages = templateFile.getPxImageFiles();
            pxFiles = memoryPxFiles;
            biFile = templateFile;
            outFilename = templateFile.getFilePath().toString();
        }

        if (isNew) {
            outFilename = String.format("%s/%s.ntpl", ET_TEMPLATE_FOLDER, templateName);
            FilePath path = new FilePath(outFilename);
            biFile = BFileSystem.INSTANCE.makeFile(path);
        }

        if (biFile == null){
            return;
        }

        try (ZipOutputStream zout = new ZipOutputStream((biFile).getOutputStream())) {
            ZipEntry zman = new ZipEntry("template-manifest.xml");
            zman.setComment("Template Manifest");

            zout.putNextEntry(zman);
            TemplateManifest manifest = config.getManifest();
            ManifestXMLWriter manifestWriter = new ManifestXMLWriter(zout);
            manifestWriter.encode(manifest);
            zout.closeEntry();
            ZipEntry zbog = new ZipEntry("template.bog");
            zbog.setComment("BOG snippet");
            zout.putNextEntry(zbog);
            ValueDocEncoder encoder = new ValueDocEncoder(zout);
            encoder.encodeDocument(root);
            zout.closeEntry();

            zipPxResources(zout, image, pxFiles, pxImages, templateName);
        }

        EasyTemplatingLogger.info(String.format(ET_OK_NTPL_CREATION_W_PARAM, outFilename));
    }

    private static void zipPxResources(
            ZipOutputStream zout,
            BImageFile image,
            PxFileRef[] pxFiles,
            BImageFile[] pxImages,
            String templateName) throws IOException {

        HashMap<String, Object> hm = new HashMap();
        int i;
        String name;

        if (pxFiles != null) {
            for (i = 0; i < pxFiles.length; ++i) {
                name = pxFiles[i].getPxName();
                hm.putIfAbsent(name, pxFiles[i]);
                BPxFile pxFile = pxFiles[i].getPxFile();

                try {
                    // Relativise PX files popupBinding to deployed PX popup file
                    EasyTemplatingUtils.addPopupBindingToObject(
                            templateName.toLowerCase(Locale.ENGLISH),
                            pxFile.getOrdInSpace(),
                            ET_POPUP_BINDING_PREFIX);
                } catch (Exception e) {
                    EasyTemplatingLogger.infoWithException("Error modifying the popup px relative ord", e);
                }

                InputStream in = pxFile.getInputStream();
                ZipEntry fileEntry = new ZipEntry("px/" + name);
                fileEntry.setTime(pxFile.getLastModified().getMillis());
                zout.putNextEntry(fileEntry);
                BajaFileUtil.pipe(in, zout);
                zout.closeEntry();
            }
        }

        if (image != null) {
            InputStream in = image.getInputStream();
            ZipEntry fileEntry = new ZipEntry("image.png");
            fileEntry.setTime(image.getLastModified().getMillis());
            zout.putNextEntry(fileEntry);
            BajaFileUtil.pipe(in, zout);
            zout.closeEntry();
        }

        if (pxImages != null) {
            hm.clear();

            for (i = 0; i < pxImages.length; ++i) {
                if (pxImages[i] != null) {
                    name = pxImages[i].getFileName();
                    zipPxImage(hm, name, pxImages[i], zout);
                }
            }
        }
    }

    private static void zipPxImage(
            HashMap hm,
            String name,
            BImageFile pxImage,
            ZipOutputStream zout) throws IOException {
        if (hm.get(name) == null) {
            hm.put(name, pxImage);
            InputStream in = pxImage.getInputStream();
            ZipEntry fileEntry = new ZipEntry("images/" + name);
            fileEntry.setTime(pxImage.getLastModified().getMillis());
            zout.putNextEntry(fileEntry);
            BajaFileUtil.pipe(in, zout);
            zout.closeEntry();
        }
    }

    private static String getUniqueName(String templateName, BDirectory dir) {
        return ObjectUtil.generateUniqueName(templateName, new NameContainerImpl(dir));
    }

    public static class NameContainerImpl implements ObjectUtil.NameContainer {
        private BDirectory dir;

        NameContainerImpl(BDirectory dir) {
            this.dir = dir;
        }

        public boolean contains(String name) {
            for (BIFile f : this.dir.listFiles()) {
                String fnWithExt = f.getFileName();
                int dot = fnWithExt.lastIndexOf('.');
                String fn = fnWithExt.substring(0, dot > 0 ? dot : fnWithExt.length());
                if (fn.equals(name)) {
                    return true;
                }
            }
            return false;
        }
    }
}
